//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./include/dollhouse/net/dolldrop.hpp
// -Declarations of the baseline RW Dolldrop sub-library

#pragma once

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "dollnet_util.hpp"
#include "mailstop_util.hpp"

//////////////////
//  'Net Work :
//

// DESIGN:
// -As a simple, basic control mechanism for starters we'll use some
// protected/private classes and a few functions from here within Dolldrop.
// Later, we can add more access restrictions and some further data validations;
// improving things bit-by-bit as we go along, using just a few of the basic &
// straightforward language features readily available to us from the standard.

// DESIGN:
// - add work-stealing queues of the Dolldrops/Mailboxes ?
// - use S.Parent's libdispatch-based async
// - this is actually a general need, so add it to the RW Foundations instead

// DESIGN:
// void push_drop() {}
// void yank_drop() {}

// DESIGN: TODO:
// return do_atomiconn(
//   lockout();
//     connect();
//       do_req(uri);
//     disconnect();
//   unlockout();
// );

// DESIGN: during this prototyping phase, code-envy is acceptable. however, for
// the production code abstract common behaviours into modular functions.
// -DRY principle

namespace rw::dollhouse::net {

fs::path const dropdir{"./dolldrops"};

//------------------------------------------------------------------------------
/**
 */
class Dolldrop_b : public RW_slave {
 public:
  Dolldrop_b(const Hip& tag, std::string const& handle);  // tag-parm'd ctor

 protected:
  /**
   * @param tag
   * @param uri
   * @return
   */
  auto chkd_pull_g(Hip const& tag, std::string const& uri) -> std::fstream;

  /**
   * @param tag
   * @param uri
   * @return
   */
  auto pull_hdr_g(Hip const& tag, std::string const& uri) -> std::stringstream;

  /**
   * @param tag
   * @param uri
   * @return
   */
  auto pull_body_g(Hip const& tag, std::string const& uri) -> std::fstream;

  /**
   * @param tag
   * @param uri
   * @return
   */
  auto pull_hdr_body_g(Hip const& tag, std::string const& uri)
      -> std::stringstream;

  /** Signals a catalog pull request
   * @note: is reset upon use, within @c chkd_pull_g() function
   */
  bool is_catalog_{false};

 private:
  Dolldrop_b() = delete;  // no default ctor

  //----------------------------------------------------------------------------
  /**
   */
  class Mailstop /* DESIGN: TODO: : public RW_worker */ {
   public:
    explicit Mailstop(Hip const& tag)
        : owner_{tag}, dollnet_{owner_, "dolldrop_dollnet"} {  //

      preflight();
    }

    /**
     * @param tag
     * @param uri
     * @return
     */
    auto try_hdr(Hip const& tag, std::string const& uri) -> std::stringstream;

    /**
     * @param tag
     * @param uri
     * @return
     */
    auto try_body(Hip const& tag, std::string const& uri) -> std::fstream;

    /**
     * @param tag
     * @param uri
     * @return
     */
    auto try_hdr_body(Hip const& tag, std::string const& uri)
        -> std::stringstream;

   private:
    Mailstop() = delete;  // no default ctor

    /** Create dolldrops directory
     */
    auto mk_dropdir() -> bool;

    /** Preliminary setups for class
     */
    void preflight();

    //-------
    class Inbox {
      //
     private:
      string data_;
    } inbox_;

    //-------
    class Outbox {
      //
     private:
      string data_;
    } outbox_;

    Hip owner_;

    //--------------------------------------------------------------------------
    /**
     */
    class Dollnet : public RW_slave {
     public:
      Dollnet(Hip const& tag, std::string const& handle);  // tag-parm'd ctor

      /** Attempts an external network connection with, and header-only download
       * transfer from, @a uri
       * @param[in] tag the caller's @c Hip
       * @param[in] uri the URI to attempt a download connection with
       * @return header stream data for @a uri, if present
       * @note: internally catches a specialization of @c std::runtime_error
       * on @b curl errors (but doesn't re-throw, currently)
       */
      auto dl_hdr(Hip const&         tag,  // guarded; for Master only
                  std::string const& uri) -> std::stringstream;

      /** Attempts an external network connection with, and download transfer
       * from, @a uri. If successful, transfer's body stream for @a uri is
       * saved out to disk file ( ./dolldrops/anons_file_nnn ).
       * @param[in] tag the caller's @c Hip
       * @param[in] uri the URI to attempt a download connection with
       * @param[in/out] dropfile the @c std::ofstream to save to
       * @note: internally catches a specialization of @c std::runtime_error
       * on @b curl errors (but doesn't re-throw, currently)
       */
      void dl_body(Hip const&         tag,  // guarded; for Master only
                   std::string const& uri,
                   std::ofstream&     dropfile);

      /** Attempts an external network connection with, and download transfer
       * from, @a uri. If successful, transfer's body stream for @a uri is
       * saved out to disk file ( ./dolldrops/anons_file_nnn ).
       * @param[in] tag the caller's @c Hip
       * @param[in] uri the URI to attempt a download connection with
       * @param[in/out] dropfile the @c std::ofstream to save to
       * @return header stream data for @a uri, if present
       * @note: internally catches a specialization of @c std::runtime_error
       * on @b curl errors (but doesn't re-throw, currently)
       */
      auto dl_hdr_body(Hip const&         tag,  // guarded; for Master only
                       std::string const& uri,
                       std::ofstream&     dropfile) -> std::stringstream;

     private:
      Dollnet() = delete;  // no default ctor

    } dollnet_;

  } mailstop_{t()};  // construct this member w/ its enclosing Dolldrop as owner

};  // class Dolldrop_b

//------------------------------------------------------------------------------
/**
 */
class Dolldrop : public Dolldrop_b {
 public:
  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file
   * @param[in] tag the caller's @c Hip
   * @param[in] uri the URI to attempt a download connection with
   * @return the disk file's handle
   * @note: public wrapper for Master's use only
   */
  auto pull(std::string const& uri) -> std::fstream {
    return chkd_pull_g(master(), uri);  // basic pull() 's are checked first
  }

  /** Attempts an external network connection with, and header-only download
   * transfer from, @a uri
   * @param[in] tag the caller's @c Hip
   * @param[in] uri the URI to attempt a download connection with
   * @return the header-only response for @a uri, if present
   * @note: public wrapper for Master's use only
   */
  auto pull_hdr(std::string const& uri) -> std::stringstream {
    return pull_hdr_g(master(), uri);
  }

  // DESIGN: TODO:
  // pull_body(){}

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file
   * @param[in] tag the caller's @c Hip
   * @param[in] uri the URI to attempt a download connection with
   * @return header stream data for @a uri, if present
   * @note: public wrapper for Master's use only
   */
  auto pull_hdr_body(std::string const& uri) -> std::stringstream {
    return pull_hdr_body_g(master(), uri);
  }

  /** Assign to the *catalog pull request* signal
   * @param flag
   * @note: controls state of protected member field in @c Dolldrop_b
   */
  void do_catalog(bool const flag);

 private:
  Dolldrop() = delete;  // no default ctor

};  // class Dolldrop

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "______"

// rw_dollhouse_lib.v220326
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
