//   RW Sysutils
// ================
// ______
// -The RW Foundational Libraries

// Filename: ./include/rw/sys/system.hpp
// -Declarations of some baseline RW Sysutils features

#pragma once

#include "incls_all.hpp"

namespace rw::sys {

// constexpr bool use_stamps{false};
constexpr bool use_stamps{true};

//------------------------------------------------------------------------------
/** Names of the most important RW Foundations class types
 */
enum class Type {
  Abstract,
  Action,
  Alexandria,
  Analogy,
  Animeyes,
  Arigato,
  Artstory,
  Atashi,
  Battery,
  Beauty,
  Bones,
  Bumpmaster,
  Categories,
  Charger,
  Chibitsu,
  Cleaner,
  Cognate,
  Collections,
  Concept,
  Cooker,
  Curator,
  Design,
  Dogfood,
  Dollhouse,
  Electro,
  Elucidate,
  Face,
  Form,
  Gears,
  Graphicom,
  Gymnasium,
  Hand,
  Hearsay,
  Meido,
  Mimicry,
  Moveit,
  Pandora,
  Percept,
  Power,
  Prop,
  Scene,
  Sempai,
  Sensei,
  Shell,
  Symmetry,
  Terminal,
  Waifusearch,

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

  Dolldrop,

#endif  // __has_include(<curl_easy.h>)

  //
  Foo  // sentinel for a general/unspecified type

};  // enum class Type

//------------------------------------------------------------------------------
/** A @c RW object's unique 'tag'. It is intended as a read-only identification
 * structure for all instantiated @c RW objects. It has these data elements:
 *
 * @b Handle: *human-readable string; it's 'name-tag'*
 *
 * @b ID: *doled-number, one for each new @c RW object created by the system*
 *
 * @b Pointer: *@c RW base object's location in the computer's memory*
 *
 * @b Type: *enumeration's item specification from the most important rw::Types*
 *
 * This is a tagging mechanism intended to be suitable for help in understanding
 * the (potentially highly-complex) callgraphs devisable for *RW Foundations*
 * systems. IE, @a 'who-called-whom-and-who-answered-the-call?'.
 *
 * It also forms the basic key for function-level access-control; to help ensure
 * against unintentional code misuse. For example, to allow for member objects
 * to only receive instructions from their owning object, since that owner
 * already knows the proper way to utilize that member's features correctly.
 * @note: The type name is an acronym: @b H andle, @b I d, @b P ointer,  @a (+)
 */
struct Hip {
  std::string   hdl{"nobody"};   // sentinel for an unassigned
  unsigned long id{9'001'42};    //  "
  RW*           ptr{nullptr};    //  "
  Type          typ{Type::Foo};  //  "

  /** Deep check for an exact equivalency between @a this, and another @c Hip
   * @param other
   * @return true if an exact match
   */
  auto operator==(Hip const& other) const -> bool;
};
// struct related
struct Hdl_id {
  string        hdl;
  unsigned long id;

  Hdl_id(Hip const& hip);  // tag-parm'd ctor

  auto operator==(Hdl_id const& other) const -> bool;

  friend std::ostream& operator<<(std::ostream& os, Hdl_id const& hid) {
    return os << hid.hdl << '.' << hid.id;
  }
};

//------------------------------------------------------------------------------
/** The collection of pointers for all @c RW objects in the system
 * @note: stored-to during the object's constructor
 *
 * DESIGN: TODO: encapsulate this data within a validating-service class object
 * @note: work to ensure only the objects themselves add/rm their handles here
 */
inline vector<RW*> RW_objs;

//------------------------------------------------------------------------------
/** Our type alias for a @c void*
 */
using Address = void*;

//------------------------------------------------------------------------------
/** Treat an @c Address as a reference to a general type @a T
 * @param[in] pt the type's @c Address to take the reference to
 * @return @a pt 's reference
 */
template <class T>
T& reference_to(Address pt) {
  return *static_cast<T*>(pt);
}

//------------------------------------------------------------------------------
/** Our type alias for a @c chrono @c duration in seconds, cast as float
 */
using Secs_f = duration<float, seconds::period>;
// alias related
using Durn = duration<long, ratio<1, 1000000000>>;

//------------------------------------------------------------------------------
/** Our type alias for a standard @c system_clock @c time_point
 */
using Time = time_point<system_clock>;
// type alias related:
using Call_clk = system_clock;

//------------------------------------------------------------------------------
/** Provides a mechanism for using @c Hip 's to trace classes' member function
 * calls. Collections of @c stack<> 's of these == *RW system callgraphs*
 * data.
 * @note: @c Hip @b tag, @c string @b func_nm, @c Time @b tim
 */
struct Callstamp {
  Hip    tag;
  string func_nm;
  Time   tim;
};

//------------------------------------------------------------------------------
/** Our type alias for a stack of @c Callstamp 's
 */
using Stampstack = stack<Callstamp, vector<Callstamp>>;

//------------------------------------------------------------------------------
/** A structure for @c Hip -tagged @c Stampstack 's
 */
struct RW_stamps {
  Hip        rw;
  Stampstack stack;
};

//------------------------------------------------------------------------------
/** Our type alias for a collection of @c Hip -tagged @c Stampstack 's
 */
using Callgraph = vector<RW_stamps>;

//------------------------------------------------------------------------------
/** Our alias for value returned by various string member functions, on failure
 * @note: aka @a 'npos'
 */
constexpr auto str_end = std::string::npos;

//------------------------------------------------------------------------------
/** Test for existence of @c std::filesystem object
 * @param p
 * @param s
 * @return
 */
auto exists_fs(fs::path const& p, fs::file_status const s = fs::file_status{})
    -> bool;

}  // namespace rw::sys

//    --=^=--

// "So is it true Sempai? Are robowaifus really better than any human woman?"

// rw_foundations_lib.v220326
// Copyright (2022)
// License MIT(Expat) https://opensource.org/licenses/MIT
