// RW Bumpmaster
// ==============
// Imageboard Retrieval & Archiving Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/curator/bumpmaster/bumpmaster.cpp
// -Definitions of the baseline RW Bumpmaster sub-library

#pragma once

#include "incls_all.hpp"

namespace rw::curator::bumpmaster {

//------------------------------------------------------------------------------
void from_json(Json const& j, Thrd_data& t) {  //

  if (! j["message"].is_null())
    t.msg = j["message"];
  else
    t.msg = "";

  if (! j["subject"].is_null())
    t.subj = j["subject"];
  else
    t.subj = "";

  if (! j["threadId"].is_null())
    t.id = j["threadId"];
  else
    t.id = 0;
}

//------------------------------------------------------------------------------
void to_json(Json& j, Thrd_data const& t) {
  j = Json{{"message", t.msg}, {"subject", t.subj}, {"threadId", t.id}};
}

//------------------------------------------------------------------------------
void from_json(Json const& j, Post_data& p) {  //

  if (! j["creation"].is_null())
    p.date = j["creation"];
  else
    p.date = "";

  if (! j["message"].is_null())
    p.msg = j["message"];
  else
    p.msg = "";

  if (! j["postId"].is_null())
    p.id = j["postId"];
  else
    p.id = 0;
}

//------------------------------------------------------------------------------
void to_json(Json& j, Post_data const& p) {
  j = Json{{"creation", p.date}, {"message", p.msg}, {"postId", p.id}};
}

//------------------------------------------------------------------------------
Bumpmaster_b::Bumpmaster_b() : Bumpmaster_b{"bumpmaster_b_obj"} {}

//------------------------------------------------------------------------------
Bumpmaster_b::Bumpmaster_b(std::string const& handle)
    : RW{handle, Type::Bumpmaster} {
  stamp("Bumpmaster_b()");
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::parse_ib_uri(std::string const& uri)
    -> std::tuple<std::string, std::string> {
  stamp("parse_ib_uri()");

  string site{""}, board{""};

  regex const ex{"(http|https)://([^/ ]*)/([^ #?]*)/([^ #?]*)"};
  std::cmatch match;

  if (regex_match(uri.c_str(), match, ex)) {
    site  = string{match[2].first, match[2].second};
    board = string{match[3].first, match[3].second};
  }

  return make_tuple(site, board);
}

//------------------------------------------------------------------------------
// void Bumpmaster_b::chk_board_dir(std::string const& site_nm,
//                                  std::string const& board_nm) {
//   //
//   auto foo = site_nm;   // TEMP:
//   auto bar = board_nm;  //  "
// }

//------------------------------------------------------------------------------
auto Bumpmaster_b::bld_cat_uri(std::string const& host, std::string const& path)
    -> std::string {
  stamp("bld_cat_uri()");

  string const scheme{"https://"};
  string const page_nm{"catalog"};
  string const type{"json"};

  return {scheme + host + '/' + path + '/' + page_nm + '.' + type};
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::bld_thrd_uri(std::string const& host,
                                std::string const& path,
                                unsigned const     thrd_id) -> std::string {
  stamp("bld_thrd_uri()");

  string const scheme{"https://"};
  string const res{"res"};
  string const type{"json"};

  // DESIGN: for now we'll just adapt to Lynxchan's formats for various things
  // (like thread URIs), since our own site's servers use it. later, we'll
  // expand out to every mainline IB type using an approach mostly similar to
  // the one previously devised for BUMP software.

  return {scheme + host + '/' + path + '/' + res + '/' + to_string(thrd_id) +
          '.' + type};
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::bump_thrd_dats(Json const& cat_json)
    -> std::vector<Thrd_data> {
  stamp("bump_thrd_dats()");

  vector<Thrd_data> res;
  res.reserve(cat_json.size());

  for (auto const& [nindx, cat_thrd] : cat_json.items()) {
    res.emplace_back(cat_thrd);  // from_json() ctors Thrd_data struct, in-place
  }

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::bump_thrd_uris(Json const&        cat_json,
                                  std::string const& site,
                                  std::string const& board)
    -> std::vector<std::string> {
  stamp("bump_thrd_uris()");

  vector<string> res;
  res.reserve(cat_json.size());

  // cycle the catalog's threads
  for (auto const& [nindx, cat_thrd] : cat_json.items()) {  //

    // build the thread's JSON URI
    auto const thrd_id  = cat_thrd["threadId"];
    auto const thrd_uri = bld_thrd_uri(site, board, thrd_id);

    res.push_back(thrd_uri);
  }

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::thrd_posts(Json const& thrd_json) -> std::vector<Post_data> {
  stamp("thrd_posts()");

  vector<Post_data> res;

  auto posts = thrd_json["posts"];
  res.reserve(posts.size());

  for (auto const& [post_idx, post_dat] : posts.items()) {
    try {
      res.emplace_back(post_dat);  // will ctor a Post_data struct in place

    } catch (Json::exception const& e) {
      puts("\n\n  caught Json exception in thrd_posts()\n\n");
    }
  }

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster_b::fmt_msg(std::string const& msg) -> std::string {
  stringstream iss{msg}, oss{};

  for (string line; getline(iss, line);) {  // read lines in one at a time
                                            //
    if (line.length() > msg_width_lim) {    // this line needs linebreaks
      // transfer line string to a sstream
      stringstream liness{line};
      line.clear();

      // parse sstream into multi-lines
      //
      string word{}, curr_line{};
      while (liness >> word) {  // one whitespace-delim'd word at a time

        if (curr_line.length() >= msg_width_lim) {  // we have a linefull, save
          line += (curr_line + '\n');
          curr_line.clear();
        }

        curr_line += (word + ' ');
      }

      curr_line = curr_line.substr(0, curr_line.size() - 1);  // rm final space
      line += curr_line;  // final line pickup for the multi-line
    }

    oss << (line + '\n');
  }

  // rm final newline
  auto res = oss.str();
  res      = res.substr(0, res.size() - 1);

  return res;
}

//------------------------------------------------------------------------------
Bumpmaster::Bumpmaster() : Bumpmaster{"bumpmaster_obj"} {}

//------------------------------------------------------------------------------
Bumpmaster::Bumpmaster(std::string const& handle) : Bumpmaster_b{handle} {
  stamp("Bumpmaster()");

  // preflight();
}

//------------------------------------------------------------------------------
void Bumpmaster::preflight() {
  stamp("preflight()");

  // TBD...
}

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

// DESIGN: NOTE: RW Mailstop likely to be moved out of rw::net, over to rw::sys

namespace rw_mail = rw::dollhouse::net;

//------------------------------------------------------------------------------
auto Bumpmaster::pull_json_drop(std::string const& json_uri)
    -> std::tuple<bool, Json> {
  stamp("pull_json_drop()");

  auto json_strm = dolldrop_.pull(json_uri);

  if (! is_empty(json_strm)) {
    auto const json = fstream_2_json(json_strm);

    if (! json.empty()) {  //

      if (sz_match(json_uri)) {
        return make_tuple(true, json);  // true == valid JSON
      }
      // else
      //   cerr << "\n  file size mismatch - pull_json_drop()\n\n\n";
    }
    // else
    //   cerr << "\n  empty JSON data - pull_json_drop()\n\n\n";
  }
  // else
  //   cerr << "\n  empty file stream - pull_json_drop()\n\n\n";

  return make_tuple(false, Json{{}});
}

//------------------------------------------------------------------------------
auto Bumpmaster::pull_json_retry(std::string const& json_uri)
    -> std::tuple<bool, Json> {
  stamp("pull_json_retry()");

  auto try_cnt{2};  // make 2 pull attempts, if needed

  do {
    auto const& [valid, json] = pull_json_drop(json_uri);
    if (valid)
      return make_tuple(valid, json);  // good to go

    // rm the old dolldrop upon failure, before re-attempting another pull
    fs::path const dropname = rw_mail::get_name(json_uri);
    if (! dropname.empty())
      fs::remove(dropname);

    // blank dolldrop's droplog record data too; should trigger a re-download
    rw_mail::set_date(json_uri, "");
    rw_mail::set_name(json_uri, "");
    rw_mail::set_sz(json_uri, "");

    // adjust the display cursor back up one line
    cout << "\u008D\r" << flush;

  } while (--try_cnt);  // loop ends at 0 (after 2 tries)

  return make_tuple(false, Json{{}});
}

//------------------------------------------------------------------------------
auto Bumpmaster::top_bump_thrds(Json const&        cat_json,
                                const int          thrd_lim,
                                std::string const& site,
                                std::string const& board)
    -> std::vector<Thrd_data> {
  stamp("top_bump_thrds()");

  vector<Thrd_data> res;

  // store all thread's data (msg, subj, ID) from the catalog; in bump-order
  auto thrd_dats = bump_thrd_dats(cat_json);

  // assemble each thread's JSON page URI; in bump-order
  auto const thrd_uris = bump_thrd_uris(cat_json, site, board);

  for (auto t_idx{0}; t_idx < thrd_lim; ++t_idx) {  // for all Page 1 threads
    auto const thrd_uri = thrd_uris[t_idx];         // index to URI

    // pull thread's dolldrop
    cout << " page_uri:    " << thrd_uri << '\n';
    auto const& [valid, thrd_json] = pull_json_retry(thrd_uri);
    //
    if (! valid) {
      puts("  invalid thread JSON\n\n");
      continue;
    }

    // copy the current thread's data into the results container
    auto& thrd_dat = thrd_dats[t_idx];       // index to thread,
    thrd_dat.posts = thrd_posts(thrd_json);  // copy post's data
    //
    res.push_back(thrd_dat);
  }

  return res;
}

//------------------------------------------------------------------------------
auto Bumpmaster::sz_match(std::string const& drop_uri) -> bool {
  stamp("sz_match()");

  auto const     logd_sz  = rw_mail::get_sz(drop_uri);    // info src: header
  fs::path const dropname = rw_mail::get_name(drop_uri);  //
  auto const     fs_sz    = fs::file_size(dropname);      // info src: disk

  return fs_sz == stoul(logd_sz);
}

//------------------------------------------------------------------------------
void Bumpmaster::do_catalog(bool const flag) {
  stamp("do_catalog()");

  dolldrop_.do_catalog(flag);
}

#endif  // __has_include(<curl_easy.h>)

}  // namespace rw::curator::bumpmaster

//    --=^=--

// "______"

// rw_curator_lib.v220326
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
